(***************************************************)
(*    HTTcc Examples for the ICFP'13 submbission   *)
(*              Functional Closures                *)
(***************************************************)

From mathcomp Require Import ssreflect ssrfun ssrbool ssrnat eqtype seq div.
From Kont Require Import rels prelude heaps domains hprop callcc srules.
Require Import Eqdep.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

(* Like Ping-Pong, but here the closures have functional type. This
extends the cooperation pattern with the ability to pass values from
the continuation to the closure, potentially affecting the behaviour
of both PING and PONG methods *)

Open Scope ccscope.

Parameters (X : lattice) (A: eqType) (x : ptr).

Program Definition incr ( x : ptr) : 
   SK X ( fun i => exists h, exists v: nat , i = x :-> v :+ h , 
          fun y i m => forall h v, i = x :-> v :+ h -> 
                         m = x :-> v .+1 :+ h /\ y = tt) :=
Do ( z <-- ! x ; x ::= z .+1 ).
Next Obligation.
move=> i [h][v]->; apply: bnd_read; apply: vrfD=>D; apply:val_write=> _ h1 v1.
by move/(cancel D)=>[->]_->; split=>//.
Qed.

Definition FalseQ ( X : Type) : post X := fun _ i m => False.

Section FunnierLoop.

Definition funnierP  j : pre :=
    fun i => forall h (v : nat),
             j = x :-> v :+ h -> i = x :-> v .+1 :+ h.

Definition funnierL (n : nat) (j :  heap) : Type := 
  forall p,  SK  X  ( 
                     fun i =>  exists h, exists t: nat, exists v : nat,
                       i = x :-> (t + v) :+ h /\ j = x :-> t :+ h
                         /\ p <= n /\ v + p = n,
                        fun (r : option nat -> SigSK X) i m => 
                           m = i /\  
                            forall (z : option nat),
                                if z is (Some y) then
                                  if (p == 0 ) then 
                                    CCC.conseq (spec_of (r z)) 
                                    (funnierP m) (CCC.ret_s y).2
                                    else 
                                      CCC.conseq (spec_of (r z)) 
                                    (funnierP m) (@FalseQ nat)
                                else
                                  CCC.conseq (spec_of (r z)) 
                                    (funnierP m) (CCC.ret_s 42).2).

Program
Definition funnierLoop  (w: bool) (n : nat) : 
  SK _ ( fun i => exists h, exists v : nat, i = x :-> v :+ h,
        fun y i m => forall h (v: nat),
                      i = x :-> v :+ h -> if w then
                      m = x :-> (v + n.+1) :+ h /\ y = n 
                        else exists q : nat,
                        m = x :-> (v + q) :+ h /\ y = 42 ) :=
  Do (c <-- callcc 
           (* R = *)
            (fun i m => 
                    forall h (v : nat), 
                  (* if n = 0 or w == false we never abort the continuation *)
                     i = x :-> v :+ h -> if (n == 0) then False else
                      exists p, 
                       m = x :-> (v + p) :+ h)
           (* Qf  *)
                ( fun j r i m => 
                    forall h (t v : nat), 
                      i = x :-> (t + v) :+ h  ->
                       j = x :-> t :+ h ->
                             m = i /\ 
                               forall (z : option nat),
                                 match z with
                                   | Some y => if (t + v == t + n) then
                                                 CCC.conseq (spec_of (r z))
                                                 (funnierP m)
                                                 (CCC.ret_s y).2
                                               else
                                                 CCC.conseq (spec_of (r z))
                                                 (funnierP m)
                                                (@FalseQ _)
                                 | None =>
                                           CCC.conseq (spec_of (r z))
                                           (funnierP m)
                                           (CCC.ret_s 42).2
                                 end)
              (fun j ex1 =>
                (ffix (fun (f: funnierL n j) (p: nat) => 
                             (Do (CCC.Match_nat p 
                                   (ret (fun q => 
                                      match q with 
                                        | Some y => BigSK (ret y)
                                        | None => BigSK (ret 42)
                                      end))
                                   (fun p' =>  
                                     ret (fun q =>
                                       match q with
                                         | Some _ => BigSK (abort _ ex1 (f p'))
                                         | None => BigSK (ret 42)
                                       end))))) n));
      z <-- ! x ;
      x ::= z .+1 ;;
      If w Then
        comm (c (Some n))
        Else
          comm (c None)).
Next Obligation. 
(* We follow the structure of the argument to ffix, thus we first case on n *)
move=> i[h][t][v] [->][J]; case: n ex1 f => [|n0] ex1 f.
(* if n = 0 then p = 0 *)
- rewrite leqn0; case; case: eqP=>//[->]_; rewrite addn0=>-> ; apply: val_ret.
 by split=>//;case => [ y|] k; rewrite addn0;
 move/(_ h t (erefl _))=>->; apply: val_ret.
(* if not, we case on p to determine on which branch we are *)
case: p=>[|p0];
(* now we case on the ifP predicates to eliminate the trivial branches*)
[rewrite addn0; move=>[_ ->]|move=>[ H H0]]; apply:val_ret; case:eqP=>//= _;
(* and then we case on z *)
(* the first two goals are trivial rets *)
split=>//; case=> [z0|] k; try (by move/(_ h (t + n0.+1) (erefl _))=>/=->);
(* and also is the last one *)
 move/(_ h (t + v) (erefl _))=>/=->; last by [].
(* The one left is the abort branch *)
apply: vrfD=>D; apply: val_abort=>/=; split=>/= [| {k} k [_ ->]].
- have D0: def (x :-> t :+ h) by rewrite !defPtUn in D *. 
 split=>//= h1 v1; rewrite J; move/(cancel D0)=>[<-]_<-.
 by exists v.+1; rewrite addnS.
 have D1 : def ( x :-> t :+ h) by (rewrite !defPtUn in D * ).
apply: val_do=>//= [|r k1 Dk [->]].
- by exists h, t, v.+1; rewrite addnS leqW// -H0 addSn addnS.
(* we case on the ifP *)
case: eqP=>// {ex1 f} H2 H3 h1 t1 v1 ; rewrite J=> H6;
move/(cancel D1)=>[HH] _ _; split=>// z; rewrite HH in H3 H6 D *;
move: (cancel D H6); case; rewrite -addnS in H3 *; move/addnI=><- _ _;
[ rewrite H2 addn1 in H0 | ] ; rewrite -H0;
(* then case on z and the hypothesis H3, this eliminate the ret goals *)
case: z (H3 z)=>// z1;case: eqP=>//.
(* Last, we proyect False out of the triple in the hypothesis*)
move=> _ {H3 k k1 Dk} H3 k P2; apply: vrfD=> Dk; move: (H3 k P2 Dk)=>/= [H4 H5].
by split=>//= y m; move/H5=> {H6} H6;move/H6=>[F]; apply: False_ind.
Qed.
Next Obligation. 
move => i[h][v]->; apply:vrfD=>D; apply: bnd_cccR=>//=.
(* qe exit i.e. first time out*)
- have P1: funnierP (x :-> v :+  h) (x :-> v .+1 :+ h).
  - by move=> h2 v2; move/(cancel D)=>[<-]_->.
 have D1 :def (x :-> v .+1 :+ h) by rewrite !defPtUn in D *.  
 apply: bnd_do=>//=; first by exists h, v, 0; rewrite addn0 add0n.
 by  move=> m k [->] H; apply: bnd_read; apply: bnd_write; case: ifP=>// W;
(* If W I'm passing some Value n, *) 
  [ case: n H=>[|n0] H;  
(* if n = 0 I'll just return 0 *)
 [ case: (H (Some 0) ( x :-> v.+1 :+ h) P1 D1)=>[H2 H3] |
(* IF n <> 0 I' ll enter the loop thus, the postcondition in the H3 hypothesis *)
(*will be False *) 
   case: (H (Some n0.+1) ( x :-> v.+1 :+ h) P1 D1)=>[H2 H3] ] |
(* if \neg W then the program will just return 42 *) 
  case: (H  None ( x :-> v .+1 :+ h) P1 D1)=>[H2 H3]]; 
 apply: val_do=>// w0 k2 D2; move/H3/(_ D2)=>[-> ->]h1 v1;
 [ rewrite addn1 | ]; move/(cancel D)=>[<-]_<- ; [| exists 1; rewrite addn1].  
(* R ; qf exit *)
(* Now we should check what happens the last time out of the loop *)
move=> k; move/(_ h v (erefl _)).
(* If n = 0 there was no throws at all *)
case:eqP=>//H0[p]->;
apply:bnd_do=>//= m k0; move/(_ h v p (erefl _) (erefl _))=>[->] H;
apply:bnd_read; apply:bnd_write; apply:vrfD=>D1.
- have P1: funnierP (x :-> ( v + p) :+ h)  (x :-> (v + p).+1 :+ h).
 move=> h1 v1.
 have _D : def (x :-> (v + p) :+ h) by rewrite !defPtUn in D *.
 by move/(cancel _D)=>[<-]_<-.
(* If W then I'll run m with Some value n else with none *)
case: ifP=>// W;  last first.
(* the (m None) case is always ret 42 *)
- case: (H None (x :-> ( v + p).+1 :+ h) P1 D1)=>[ P2 H3].
 apply: val_do=>// w0 k1 Dk1; move/(H3 w0 k1)/(_ Dk1) =>[-> ->] h1 v1. 
 by case/(cancel D)=>-> _ ->; exists p.+1; rewrite addnS.
(* if I'm in the last iteration, i.e.  v + p = v + n I'll, I reached the end. *)
(* This means that (m (Some n)) just returns n . If not, it will jump, thus *)
(* making Q_m False *)
move: (H (Some n)); case: eqP=>// ; [ move/addnI=> H1 | move=> H1 ];
 move/(_ (x :-> (v + p).+1 :+ h) P1 D1) => [P2 H3]; 
 apply: val_do=>// w0 k1 Dk1; move/(H3 w0 k1)/(_ Dk1) =>[-> ->] h1 v1. 
 by move/(cancel D)=>[-> _ ->]; rewrite H1 addnS.
Qed.
End FunnierLoop.
