(***************************************************)
(*    HTTcc Examples for the ICFP'13 submbission   *)
(*                     incr3                       *)
(***************************************************)

From mathcomp Require Import ssreflect ssrfun ssrbool ssrnat eqtype seq div.
From Kont Require Import rels prelude heaps domains hprop callcc srules.
Require Import Eqdep.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Open Scope ccscope.
Parameter X: lattice.

(* Incr from Example 3 in Section 3 *)

(* We unfold the defininiton of separation star (#), this, \In, etc
and deal with explicit heap equality and heap union :+ :

 p1 # p2 : Pred heap :=
Pred_of_Simpl
  [Pred h | exists h1 h2 : heap, h = h1 :+ h2 /\ h1 \In p1 /\ h2 \In p2]

this h  : Pred heap := Pred_of_Simpl [Pred h | i = h]

i \In P <-> P i

Other relevant definitions and properties can be found in the heaps.v
and hprop.v. For further details on def predicates, see

[26] NANEVSKI, A., VAFEIADIS, V., AND BERDINE, J. Structuring the
verification of heap-manipulating programs. In POPL (2010). *)


Program Definition incr ( x : ptr) : 
   SK X ( fun i => exists h, exists v: nat , i = x :-> v :+ h , 
          fun y i m => forall h v, i = x :-> v :+ h -> 
                         m = x :-> v .+1 :+ h /\ y = tt) :=
Do ( z <-- ! x ; x ::= z .+1 ).
Next Obligation.     
move=> i [h][v]->; apply: bnd_read; apply: vrfD=>D; apply:val_write=> _ h1 v1.
by move/(cancel D)=>[->]_->; split=>//.
Qed.

Definition incr3P1 x v  h : pre :=
     fun i => i = x :-> v .+1 :+ h. 

Definition incr3P2 x v  h : pre :=
     fun i => i = x :-> v .+3 :+ h. 

Definition incr3Q2 (x : ptr) : post unit := (CCC.ret_s tt).2.

Program
Definition incr3 (x: ptr) : SK X 
  ( fun i =>  exists v : nat, exists h, i = x :-> v :+ h,
     fun y i m =>  forall v h, i = x :-> v :+ h ->
                       m = x :-> v .+3 :+ h /\ y = tt) :=
  Do (c <-- @CCC.callcc X (SigSK X)
(* R = *)
                (fun i m =>  forall h (v : nat), 
                              i = x :-> v  :+ h ->
                              m = x :-> v .+1  :+ h)
(* Qf *)         
          (fun j z i m => forall h (v: nat),
                                j = x :-> v :+ h ->
                                i = x :-> v .+1 :+ h ->
                                m = x :-> v .+2 :+ h  /\
                               CCC.conseq (spec_of z) (incr3P2 x v h) 
                                                      (incr3Q2 x))
(* se j *)
                   (fun j =>
                     (fun i =>
                       exists v : nat, exists h, i = x :-> v :+ h /\ j = i
                       , fun z i m => forall v h,
                                  i  = x :-> v :+ h -> j = i ->
                              m = i /\
                             CCC.conseq (spec_of z) 
                             (incr3P1 x v h ) (fun r i m => False)))
                                
        (fun j ex1 =>
           Do (ret (BigSK (abort _ ex1 (incr x ;; ret (BigSK (ret tt)))))));
      incr x ;;
      comm c).
Next Obligation.
rewrite !gh; apply:ghE; first by move=> i [v][h][I J]_; exists (v, h).
move=> i s [H0]_[v][h][I J]; rewrite I in H0 J *=>D. 
case:(cancel D H0)=><- _ <-{H0}; apply: val_ret;split=>//= k. 
rewrite J=>->.
apply: val_abort; split=>//=; first by split=>// h2 v2;move/(cancel D)=>[<-]_->.
move=>/={k I}k[_ ->]; apply:bnd_do=>//=; first by hhauto.
case=> {k J}k; move/(_ h (v.+1) (erefl _))=>[->] _.
apply:val_ret=>//h1 v1; move/(cancel D)=>[<-]_<- _.
by split=>//= {k}k; move/(_ h v (erefl _))=>->; apply: val_ret.
Qed.
Next Obligation.
rewrite !gh; apply:ghE; first by move=> i [v][h]-> _; exists (v, h).
move => i s H0 [v][h]I; rewrite I in H0 *=> D.
case:(cancel D H0)=> <- _ <- {s H0 I}; apply: bnd_ccc.
+ apply: bnd_do=>//= ; first by hhauto.
 move => c k; case/(_ v h (erefl _) (erefl _))=> -> H2.
 apply: bnd_do; first by hhauto.
 move=> {k} tt k; move/(_ h v (erefl _))=>[->]_; apply: vrfD=>D1.
 have P1: incr3P1 x v h (x :-> v.+1 :+ h)  by [].
 case: (H2 (x :-> v.+1 :+ h) P1 D1)=>[H3 H4]. 
 apply:val_do=>// w k1 Dk1 P.
 by move:( H4 w k1 P Dk1); apply: False_ind. 
move=> m k{D} [h1][_];case; move/(_ h v (erefl _))=>->.
 case/(_ h v (erefl _) (erefl _)) =>-> H1.
apply: vrfD=>D1;apply: bnd_do; first (by hhauto); move=>/= {k h1} tt k.
move/(_ h v.+2 (erefl _))=>[->]_.
have P2: incr3P2 x v h (x :-> v.+3 :+ h) by [].
apply:vrfD; case/(H1 (x :-> v.+3 :+ h) P2)=>[H3 H4]. 
apply: val_do=>//= {k tt} tt k Dk P.
by case:( H4 tt k P Dk)=> -> ->. 
Qed.
