(***************************************************)
(*    HTTcc Examples for the ICFP'13 submbission   *)
(*      Simple downward continuation examples      *)
(***************************************************)

From mathcomp Require Import ssreflect ssrfun ssrbool ssrnat eqtype seq div.
From Kont Require Import rels prelude heaps domains hprop callcc srules.
Require Import Eqdep.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Open Scope ccscope.

Parameter (X: lattice).

(* Simple Downward continuation examples: escaping out of
computational context and playing a little bit with pointers *)

Section ExampleSetR.

Program
Definition exra0 : SK X (xPredT, fun y i m => i = m /\ y = 10) :=
  Do ( x <-- alloc 0 ;
       x ::= 10 ;;
       c <-- callcc (fun i m => False) 
                    (fun _  y i m => False)
                 (fun j f =>  ret 0);
       b <-- ! x ;
       dealloc x ;;
       ret (b + c)).
Next Obligation.
move => i _; apply: bnd_alloc =>x ; apply:bnd_write; apply:vrfD=>D1.
apply:bnd_cccR=>//=.
by apply:bnd_ret; apply:bnd_read; apply:bnd_dealloc; apply:val_ret.
Qed.   

Program
Definition exrb0 : SK X (xPredT, fun y i m => i = m /\ y = 40) :=
  Do ( x <-- alloc 0 ;
       x ::= 10 ;;
       c <--callcc (fun i m => forall h (v : nat), i = x :-> v :+ h -> m = i)
                   (fun _ y i m => 
                           forall h (v: nat),
                            i = x :-> v :+ h ->
                            m = x :-> 40 :+ h /\ y = tt)
                        (fun j f => abort _ f (x ::= 40)) ;
       b <-- ! x ;
       dealloc x ;;
       ret b).
Next Obligation.
move => i _; apply: bnd_alloc=> x; apply: bnd_write; apply: vrfD=> D0.
apply: bnd_cccR=>//=.
+ apply: bnd_abort=>//=; split; first by (hhauto). 
 move =>/= j [[H1 _]->]. (*  move:(H1 i 10 (erefl _))=>-> *)
 by apply: val_write=> D h v H; move: (cancel D0 H)=>[_]_->.
move=> j H ; move:(H i 10 (erefl _))=>->; apply: bnd_do=>//= tt s.
move/(_ i 10 (erefl _))=>[->]_ ; apply: bnd_read; apply:bnd_dealloc.
by apply: val_ret.
Qed.

Program 
Definition exrc0 : SK X (xPredT, fun y i m => i = m /\ y = 40) :=
  Do ( x <-- alloc 0 ;
       c <-- callcc (fun i m => 
                            forall h (v : nat),
                              i = x:-> v :+ h -> m = x:-> 10 :+ h )
                 (fun _ y i m => 
                      forall h (v: nat),
                           i = x :-> v :+ h -> m = x :-> 40 :+ h /\ y = tt )
                     (fun j exit =>
                                x ::= 10 ;;
                                abort unit exit (x ::= 40 ) ;;
                                x ::= 100);
       b <-- ! x ;
       dealloc x ;;
       ret b).
Next Obligation.
Proof.
move => i _; apply: bnd_alloc=> x; apply: vrfD=> D0.
apply:bnd_cccR=>//=.
+ apply:bnd_bnd; apply:bnd_write; apply:bnd_bnd; apply:bnd_abort=>/=.
 split=>//=; first by split=>//h v; move/(cancel D0)=>[_ _ ->]. 
 move=> j [[H1 _]]->. (*; move:(H1 i 0 (erefl _))=>->;  *)
 by apply: vrfD=> D1;apply: val_write=> _ h v; move/(cancel D1)=>[_]_->. 
move =>/= j; move/(_ i 0 (erefl _))=>->.
apply: bnd_do=>//= tt k; move/(_ i 10 (erefl _))=>[->]_.
by apply: bnd_read; apply:bnd_dealloc; apply:val_ret.
Qed.

Program 
Definition exRd0 : SK X (xPredT, fun y i m => i = m /\ y = 50) :=
  Do ( x <-- alloc 0 ;
       c <-- callcc (fun i m => 
                           forall h (v: nat),
                             i = x :-> v :+ h -> m = x :-> 10 :+ h)
                    (fun _ y i m => 
                           forall h (v: nat),
                             i = x :-> v :+ h -> m = x :-> 40 :+ h /\ y = v)
                     (fun j exit => 
                                x ::= 10 ;;
                                abort nat exit (y <-- ! x;
                                            x ::= 40;;
                                            ret y) ;;
                                x ::= 100 ;;
                                ret 0);
       b <-- ! x ;
       dealloc x ;;
       ret (b + c)).
Next Obligation.
move=>i _; apply:bnd_alloc=>x; apply:vrfD=>D0.
apply:bnd_cccR=>/=.
+ apply:bnd_bnd; apply:bnd_write; apply:vrfD=>D1; apply:bnd_bnd.
 apply: bnd_abort=>/=;split; first by split=>//h v; move/(cancel D0)=>[_ _ ->].
 move=>j[[H1 _]]->;  apply:bnd_read; apply: bnd_write.
 by apply: val_ret=>h v; move/(cancel D1)=>[<-]_->.
move => k; move/(_ i 0 (erefl _))=>->; apply: vrfD=>D1.
apply:bnd_do=>//= w s; move/(_ i 10 (erefl _))=>[-> ->].
by apply: bnd_read; apply:bnd_dealloc; apply:val_ret.
Qed.

(* Swapping and swapping back *)

Program 
Definition exRe0 : SK X (xPredT, fun y i m => i = m /\ y = 100) :=
  Do ( x <-- alloc 0 ;
       y <-- alloc 100 ;
       c <-- callcc (fun i m =>
                              forall h (v w: nat),
                                i = y :-> v :+ x :-> w :+ h ->
                                m = y :-> w :+ x :-> v :+ h )                           
                    (fun _ z i m =>
                              forall h (v w: nat),
                                i = y :-> v :+ x :-> w :+ h ->
                                m = y :-> w :+ x :-> v :+ h /\ z = tt)
                     
                           (fun j exit =>
                                a <-- @CCC.read X nat x ;
                                b <-- @CCC.read X nat y ;
                                x ::= b ;;
                                y ::= a ;;
                                abort unit exit ( y ::= b ;; x ::= a) ;;
                                ret tt) ;
       b <-- ! y ;
       dealloc x ;;
       dealloc y ;;
       ret b ).
Next Obligation.
move=>i _; apply:bnd_alloc=>x; apply:bnd_alloc=>y; apply:vrfD=>D0.
apply:bnd_cccR=>//=.
+ do 2 (rewrite unCA; apply: bnd_bnd; apply: bnd_read).
  do 2 (rewrite unCA; apply: bnd_bnd; apply: bnd_write); apply: bnd_bnd.
  apply: vrfD=> D1; apply: bnd_abort=>/=; split=>//=.
  - split=>// h v w; rewrite unCA -!unA;move/(cancel D0)=>[->]D2. 
   by move/(cancel D2)=>[-> _ ->].
  move=>/= j[[H _]]->;rewrite unCA unA in H.
  apply: bnd_write; rewrite unCA; apply: val_write=> _ h v w.
  rewrite -!unA; move/(cancel D1)=>[<-]D2; move/(cancel D2)=>[-> _ ->].
  by rewrite unCA.
move=> j; rewrite unA;move/(_ i 100 0 (erefl _))=>->; apply: bnd_do=>//= tt k.
move/(_ i 0 100 (erefl _))=>[->]_; rewrite -unA.
by apply:bnd_read; rewrite -unCA; do 2 apply:bnd_dealloc; apply: val_ret.
Qed.

(* A callcc/fix example from the older example set *)

Definition ex6T w j x : Type := 
  forall p: nat , SK X ( fun i => 
                             exists z : nat, exists h, 
                                   i = x :-> z :+ h /\ 
                                    j = x :-> 0 :+ h  /\ z + p = w,
                        fun (y: nat) i m => 
                           forall h (v : nat),
                             i = x :-> v :+ h ->
                          if ( w > 4) then False else
                             m = x :-> (v + p) :+ h /\ y = 0).

Program
Definition ex6 w : SK X (xPredT, fun y i m => i = m /\ 
                               if w > 4 then y = w + 1 else y = w ) :=
  Do ( x <-- alloc 0 ;
       n <-- alloc w ; 
       c <-- callcc  (fun i m => if (w > 4) then
                                     forall h (v: nat), 
                                       i = x :-> v :+ h ->   
                                          m = x :-> w :+ h
                                    else False)
                     ( fun j y i m => (CCC.ret_s 1).2 y i m)
                     (fun j exit => 
                       b <-- ! n ; 
                       c <-- ffix (fun (f: ex6T w j x ) (y: nat) => 
                             Do (CCC.Match_nat y (If (w > 4) 
                                                  Then (abort _ exit (ret 1))
                                                  Else (ret 0)) 
                                             (fun z => p <-- ! x ;
                                              x ::= p .+1 ;;
                                              f z))) b;
                       ret 0);
       b <-- ! x ;
       dealloc x ;;
       dealloc n ;;
       ret (b + c)).
Next Obligation.
move=>k[z][h][->][->];case:y=>[E|y E]; apply:vrfD=>D0.
+ rewrite addn0 in E; rewrite E in D0 *; case:ifP=>//=[->|P]. 
 - apply:val_abort=>/=; do 2 split=>//; move=> s p.
  have D1: def (x:-> 0 :+ h) by rewrite !defPtUn in D0*.
  by case/(cancel D1)=>_ _ ->.
 by apply:val_ret=> s v; rewrite addn0.
apply:bnd_read; apply: bnd_write; apply:val_do=>//=.
+ by exists z.+1; exists h; rewrite addnS in E; rewrite -E.
move=>q s Ds H i p; case/(cancel D0)=><-_ <-.
by move:( H h z.+1 (erefl _)); case: ifP=>//H0[-> ->]; rewrite addSn addnS.
Qed.
Next Obligation.
move => i _; apply: bnd_alloc=> x; apply: bnd_alloc=> n.
apply: bnd_cccR =>/= .
+ apply: bnd_bnd; apply: bnd_read; apply: bnd_bnd; apply: bnd_do=>//=.
  - exists 0; exists ( n :-> w :+ i).
   by rewrite unCA; hhauto; case: ifP=>//_; exists 0.
  move=>z j; rewrite unCA; move/(_ (n :-> w :+ i) 0 (erefl _)). 
  case: ifP=>// _ [->]_; rewrite add0n; apply: bnd_ret.
  by apply: bnd_read; do 2 apply: bnd_dealloc; apply: val_ret.
move=> j; rewrite unCA; case: ifP=>// _.
move/(_ (n :-> w :+ i) 0 (erefl _)) ->.
apply: bnd_do=>//= z k[-> ->]; apply: bnd_read; do 2 apply: bnd_dealloc.
by apply: val_ret.
Qed.

Definition ex7T (w: nat) j x : Type := 
  forall p: unit , SK X ( fun i => exists h, exists v: nat,
                                  i = x :-> v :+ h /\
                                  j = x :-> 0 :+ h , 
                        fun (y: unit) i m => False).                        

Program
Definition ex7 w : SK X (xPredT, fun y i m =>  i = m /\ y = w)
                                :=
  Do ( x <-- alloc 0 ;
       c <-- callcc (fun i m => forall h (v: nat),
                                 i = x :-> v :+ h -> m = x :-> w :+ h)
                      (fun _ y i m => (CCC.ret_s tt).2 y i m)
                     (fun j exit => 
                       c <-- ffix (fun (f: ex7T w j x ) (y: unit) => 
                                           Do (
                                              p <-- @CCC.read X nat x ; 
                                              If  (w == p) Then 
                                                   (CCC.throw exit tt)
                                              Else
                                              (x ::= p .+1 ;;
                                              f y))) tt;
                       ret c);
       b <-- !x ;
       dealloc x ;;
       ret b).
Next Obligation.
move=>/= i [h][v][-> ->].
apply:bnd_read; case:eqP=>/= H1; apply: vrfD=>D0; last first.
- by apply:bnd_write; apply: val_do=>//=; hhauto.
apply:val_throw=>//=.
split=>//; have D1: def (x :-> 0 :+ h) by rewrite !defPtUn in D0 *.
by move=>/=  h1 v1 ; case/(cancel D1)=>_ _ ->; rewrite H1.
Qed.
Next Obligation.
move =>/= i _;apply: bnd_alloc=>/= x; apply: bnd_cccR=>//.
+ by apply: bnd_bnd; apply: bnd_do=>//=; first by hhauto.
move=> j; move/(_ i 0 (erefl _))=>->; apply: bnd_do=>//= tt s[_]->.
by apply: bnd_read; apply: bnd_dealloc; apply: val_ret.
Qed.
End ExampleSetR.

(* Nesting callcc's: One step exit vs top exits *)

Definition RTrue : heap -> heap -> Prop  := fun k => xPredT. 

Definition RFalse : heap -> heap -> Prop := fun k => xPred0.

Definition RSame : heap -> heap -> Prop := xPred1.

Definition freespec X : heap -> spec X := fun _ => (xPredT, fun y => RTrue ). 

Section ExampleSet2.

Variable (A: Type).

Program
Definition ex21 : SK X (xPredT, fun y i m =>  i = m /\ y = 0) :=
Do (x <-- alloc 0 ;
    c <-- callcc (fun i m => False) (fun _  y i m => False)
        (fun j ex1 => 
           y <-- !x ; 
           x ::= y .+1 ;;
           d <-- callcc xPred1
                (fun _ y i m => forall h v, 
                        i = x :-> v :+ h ->
                        m = x :-> v .+1 :+ h /\ y = tt)
                   (fun k ex2 => abort _ ex2 ( y <-- !x ;
                                               x ::= y .+1;;
                                               ret tt));                  
(* We've made a one step exit, so the following code will run and reset x *)
           x ::= 0 ;;
           ret tt);
    z <-- ! x ;
    dealloc x ;;
    ret z).
Next Obligation.
move => j _; apply: bnd_alloc=>x; apply: bnd_cccR=>//.
apply: bnd_bnd; apply: bnd_read; apply: bnd_bnd; apply: bnd_write.
apply: vrfD=>D0; apply: bnd_bnd; apply: bnd_cccR=>/=.
+ apply: bnd_abort=>/=; split=>//; move=> i [[<-] _]_ ; apply: bnd_read.
 by apply: bnd_write; apply: val_ret=> h v; case/(cancel D0)=><-_->.
move=> k <-; apply: bnd_do=>//= tt h; case/(_ j 1 (erefl _))=>-> _.
apply: bnd_bnd; apply: bnd_write; apply: bnd_ret; apply: bnd_read.
by apply: bnd_dealloc; apply: val_ret.
Qed.

(* Now, we will abort from the inner level with the outer level continuation: *)
(* this should ignore the code after the inner callcc call in the outer block *)

Program
Definition ex22 : SK X (xPredT, fun y i m =>  i = m /\ y = 2) :=
Do (x <-- alloc 0 ;
    c <-- callcc (fun i m => forall h v,
                        i = x :-> v :+ h ->
                        m = x :-> v .+1 :+ h)
                 (fun _ y i m => forall h v, 
                        i = x :-> v :+ h ->
                        m = x :-> v .+1 :+ h /\ y = tt)
        (fun j ex1 => 
           y <-- !x ; 
           x ::= y .+1 ;;
           d <-- callcc (fun _ => xPred0)
                     (fun _  _ _ => xPred0)                
                   (fun k ex2 => abort unit ex1 ( y <-- !x ;
                                               x ::= y .+1;;
                                               ret tt));                  
           x ::= 0 ;;
           ret tt);
    z <-- ! x ;
    dealloc x ;;
    ret z).
Next Obligation.
move => j _; apply: bnd_alloc=>x; apply: vrfD=>D0; apply: bnd_cccR=>//.
apply: bnd_bnd; apply: bnd_read; apply: bnd_bnd; apply: bnd_write.
apply: vrfD=>D1; apply: bnd_bnd; apply: bnd_cccR=>//=.
+ apply: bnd_abort=>//=; split.
 - by split=>// h v; case/(cancel D0)=>->_->.
 move => h [[H _]]->.
 apply:bnd_read; apply:bnd_write; apply:val_ret=> k w.
 by case/(cancel D1)=>->_->.
move => k; move/(_ j 0 (erefl _))=>->; apply: bnd_do=>//=.
move=>/= tt k1; case/(_ j 1 (erefl _))=>[->]_.
by apply: bnd_read; apply: bnd_dealloc; apply: val_ret.
Qed.

(* Reentry attempts *)
(* We know we cannot return the same continuation, not throw a continuation *)
(* to itself, or store a continuation in the heap -- w/o moving to an *)
(* impredicative setting. But there' s still some options*)

(* We will try to return some other STcont. The only wsay we've got to create *)
(* continuations is by nesting callcc calls *)


(* Testing returning and aborting computations *)
(* In Bauer and Pretnar, yield and fork have resp () -> () and *)
(* ( () -> () ) -> () types. This might imply we do not have to take big *)
(* computation types as arguments, rather pass around small values on which *)
(* a computation may act later i.e returning a monadic function A --> ST p B q *)
(* which is deffered until later. *)

(* Returning a function : () -> p ST q *)

Program
Definition ex23a : SK X (xPredT, fun y i m =>  i = m /\ y = 1) :=
Do (x <-- alloc 0 ;
    c <-- callcc RFalse (fun _ y => RTrue)
        (fun j ex1 => 
           x ::= 2 ;; 
          ret (fun (k: unit) => x ::= 1));
    w <-- (c tt) ;
    z <-- ! x ;
    dealloc x ;;
    ret z).
Next Obligation.
move => j _; apply: bnd_alloc=> x ; apply: vrfD => D0; apply: bnd_cccR=>//.
apply: bnd_bnd; apply:bnd_write; apply: bnd_ret; apply: bnd_write; apply: bnd_read.
by apply: bnd_dealloc; apply: val_ret.
Qed.

(* A nicer one *)

Program
Definition ex23b : SK X (xPredT, fun y i m =>  i = m /\ y = 0) :=
Do (x <-- alloc 0 ;
    c <-- callcc RFalse (fun _ y=> RTrue) 
        (fun j ex1 => 
           x ::= 2 ;; 
          ret (fun (n: nat) => x ::= n));
    w <-- (c 0) ;
    z <-- ! x ;
    dealloc x ;;
    ret z).
Next Obligation.
move => j _; apply: bnd_alloc=> x ; apply: vrfD => D0; apply: bnd_cccR=>//.
apply:bnd_bnd; apply:bnd_write; apply: bnd_ret; apply: bnd_write; apply: bnd_read.
by apply: bnd_dealloc; apply: val_ret.
Qed.
End ExampleSet2.

