(**************************************************)
(* HTTCC Examples from the  manuscript            *)
(*   LARGE-FOOTPRINT SEMANTICS VERSION            *)
(*  call make clean; make WH=1  first             *)  
(**************************************************)

From mathcomp Require Import ssreflect ssrfun ssrbool ssrnat eqtype seq. 
From Kont Require Import rels prelude heaps hprop callccWH srulesWH.
Require Import Eqdep.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Open Scope ccscope.

Program Definition incr ( x : ptr) : 
   ST ( fun i => exists h, exists v: nat , i = x :-> v :+ h , 
          fun y i m => forall h v, i = x :-> v :+ h -> 
                         m = x :-> v .+1 :+ h /\ y = tt) :=
Do ( z <-- ! x ; x ::= z .+1 ).
 (* Obligation 1: *)
 (* x : ptr *)
 (*  ============================ *)
 (*   forall i : heap, *)
 (*   (exists h : heap, exists v : nat, i = x :-> v :+ h) -> *)
 (*   CCC.verify i *)
 (*     (CCC.bind_s (CCC.read_s nat x) (fun z : nat => CCC.write_s x z.+1)) *)
 (*     (fun (y : unit) (m : heap) => *)
 (*      forall (h : heap) (v : nat), *)
 (*      i = x :-> v :+ h -> m = x :-> v.+1 :+ h /\ y = tt) *)
Next Obligation.     
move=> i [h][v]->; apply: bnd_read; apply: vrfD=>D; apply:val_write=> _ h1 v1.
by move/(cancel D)=>[->]_->; split=>//.
Qed.

Section Twice.

Definition twiceP1 j (x : ptr)  : pre :=
     fun i => forall h (v : nat), 
                 j = x :-> v :+ h -> 
                    i = x :-> v .+1 :+ h. 

Definition twiceP2 j (x : ptr) : pre :=
     fun i => forall h (v : nat), 
                 j = x :-> v :+ h -> 
                    i = x :-> v .+2 :+ h. 

Definition twiceQ1: post unit := fun y i m => False.
Definition twiceQ2 (x : ptr) : post unit := (CCC.ret_s tt).2.

Program
Definition twice (x: ptr) : ST 
  ( fun i =>  exists v : nat, exists h, i = x :-> v :+ h,
     fun y i m =>  forall v h, i = x :-> v :+ h ->
                       m = x :-> v .+2 :+ h /\ y = tt) :=
  Do (c <-- @CCC.callcc _
(* sf *)
                (fun i => exists h, exists v : nat, 
                              i = x :-> v  :+ h, 
                fun r i m => forall h (w: nat),
                                i = x :-> w  :+ h ->
                                ( m = i /\ CCC.conseq (spec_of r) 
                             (twiceP1 i x ) twiceQ1)
                                \/ 
                                ( m = x :-> w.+1 :+ h  /\
                               CCC.conseq (spec_of r) (twiceP2 i x) (twiceQ2 x)))

                   (fun j =>
                     (fun i =>
                       exists v : nat, exists h, i = x :-> v :+ h /\ j = i
                       , fun z i m => 
                              m = i /\
                             CCC.conseq (spec_of z) 
                             (twiceP1 j x ) twiceQ1))
                                
        (fun j ex1 =>
           Do (ret (BigST (throw ex1 (BigST (ret tt)) _)))) _;
      incr x ;;
      comm c).
Next Obligation.
move=> i [v][h][->]J; apply: vrfD=> D; apply: val_ret; split=>//= k.
move/(_ h v J)=>->; apply: val_throw=> h1 v1.
rewrite J; case/(cancel D)=><- _ <-; right; split=>//{k}k.
Qed.
Next Obligation.
rewrite !gh; apply: ghE; first by move=> i [_] [v][h]-> _; exists (v,h).
move=> i p -> [->][h][v] E; rewrite E=>D; apply: val_do=>//=; first by hhauto.
move=> m k Dk [->] H2.
have P1: twiceP1 (x :-> v :+ h) x (x :-> v.+1 :+ h). 
 - by move=> h1 v1 ;case/(cancel D)=>->_->.
have D1: def (x :-> v.+1 :+ h) by rewrite !defPtUn in D *.
by left; case: (H2 (x :-> v.+1 :+ h) P1 D1)=>[H3 H4]; split=>//.
Qed.
Next Obligation.
rewrite !gh; apply: ghE; first by move=> i [v][h]-> _; exists (v,h).
move=> i p -> [v][h]E; rewrite E=> D; apply: bnd_do=>// ; first by hhauto.
move=>/= m k; case/(_ h v (erefl _))=>[[->] H2|[->H2]].
(* First iteration*)
+ apply:bnd_do; first (by hhauto); move=>/= tt {k} k .
 move/(_ h v (erefl _))=>[->]_.
have P1: twiceP1 (x :-> v :+ h) x (x :-> v.+1 :+ h). 
 - by move=> h1 v1 ;case/(cancel D)=>-> _->.
have D1: def (x :-> v.+1 :+ h) by rewrite !defPtUn in D *.
case: (H2 (x :-> v.+1 :+ h) P1 D1)=>[H3 H4]; apply:val_do=>//= {k tt} tt k Dk.
by move/(H4 tt k); move/(_ Dk); apply: False_ind.
(*jumped back in *)
(* Last iteration *)
apply:bnd_do; first by hhauto.
move=>/= tt{k} k; move/(_ h v.+1 (erefl _))=>[->]_.
have P1: twiceP2 (x :-> v :+ h) x (x :-> v.+2 :+ h). 
 - by move=> h1 v1 ;move/(cancel D)=>[->]_->.
have D1: def (x :-> v.+2 :+ h) by rewrite !defPtUn in D *.
case: (H2 (x :-> v.+2 :+ h) P1 D1)=>[H3 H4]; apply: val_do=>//= {k tt} tt k Dk P. 
by case:(H4 tt k P Dk)=>->->; case:(cancel D (esym E))=><- _ <-.
Qed.

(* (* Proof in the algebraic model*) *)

(* Next Obligation. *)
(* move=> i [v] [h][->][J]; apply: vrfD=>D0; apply: val_ret. *)
(* split=>//= i1; rewrite J; move/(_ h v (erefl _))=>[->]. *)
(* apply: val_abort; split; first by split=>// h2 v2; move/(cancel D0)=>[<-]_->. *)
(* move=>/= k [_ [->]]; apply: bnd_do=>//=; first by hhauto. *)
(* move=> tt k1; move/(_ h (v.+1) (erefl _))=>[->] _. *)
(* apply: val_ret=>// h1 v1; move/(cancel D0)=>[<-]_<- _. *)
(* by split=>//= k2; move/(_ h v (erefl _))=>->; apply: val_ret. *)
(* Qed. *)
(* Next Obligation. *)
(* move => i [v][h][->] ; apply: vrfD=> D0; apply: bnd_cccR. *)
(* + apply: bnd_do=>/= ; first by hhauto. *)
(*  move => c k [->] H2. *)
(*  apply: bnd_do; first by hhauto. *)
(*  move=> {k} tt k; move/(_ h v (erefl _))=>[->]_; apply: vrfD=>D1. *)
(*  have P1: incr3P1 (x :-> v :+ h) x (x :-> v.+1 :+ h). *)
(*  - by move=> h1 v1 ;move/(cancel D0)=>[->]_->. *)
(*  case: (H2 (x :-> v.+1 :+ h) P1 D1)=>[H3 H4].  *)
(*  apply:val_do=>// w k1 Dk1 P. *)
(*  by move:( H4 w k1 P Dk1); apply: False_ind.  *)
(* move=> j; move/(_ h v (erefl _))=>[->]. *)
(* apply: bnd_do=>//= m k; move/(_  h v (erefl _) (erefl _))=>[->]H1. *)
(* apply: vrfD=>D1;apply: bnd_do; first (by hhauto); move=>/= {k} tt k. *)
(* move/(_ h v.+2 (erefl _))=>[->]_. *)
(* have P1: incr3P1 (x :-> v.+2 :+ h) x (x :-> v.+3 :+ h). *)
(*  by move=> h1 v1; move/(cancel D1)=>[->]_[->]. *)
(* apply:vrfD; case/(H1 (x :-> v.+3 :+ h) P1)=>[H3 H4].  *)
(* apply: val_do=>//= {k tt} tt k Dk P. *)
(* by move:( H4 tt k P Dk)=>[<-][->] v1 h1; move/(cancel D0)=>[->][_][->]. *)
(* Qed. *)
End Twice.


(*  We move now to trying to do incr3: we will change abort f e to e ; *)
(* throw f to mimic the behaviour of the intermediate code *)

Section Incr3.
Definition incr3P1 j (x : ptr)  : pre :=
     fun i => forall h (v : nat), 
                 j = x :-> v :+ h -> 
                    i = x :-> v .+1 :+ h. 

(* Definition incr3P2 j (x : ptr)  : pre := *)
(*      fun i => forall h (v : nat),  *)
(*                  j = x :-> v :+ h ->  *)
(*                     i = x :-> v .+2 :+ h.  *)


Definition incr3Q1: post unit := fun y i m => False.
Definition incr3s := (CCC.ret_s tt).


Program
Definition incr3 (x: ptr) : ST 
  ( fun i =>  exists v : nat, exists h, i = x :-> v :+ h,
     fun y i m =>  forall v h, i = x :-> v :+ h ->
                       m = x :-> v .+3 :+ h /\ y = tt) :=
  Do (c <-- @CCC.callcc _
(* sf *)
                (fun i => exists h, exists v : nat, 
                              i = x :-> v  :+ h, 
                fun r i m => forall h (w: nat),
                                i = x :-> w  :+ h ->
                                ( m = i /\ CCC.conseq (spec_of r) 
                             (incr3P1 i x) incr3Q1)
                                \/ 
                                ( m = x :-> w.+2 :+ h  /\
                               CCC.conseq (spec_of r) (incr3s.1) incr3s.2))

                   (fun j =>
                     (fun i =>
                       exists v : nat, exists h, i = x :-> v :+ h /\ j = i
                       , fun z i m => 
                              m = i /\
                             CCC.conseq (spec_of z) 
                             (incr3P1 j x ) incr3Q1))
                                
        (fun j ex1 =>
           Do (ret (BigST (incr x ;; throw ex1 (BigST (ret tt)) _)))) _;
      incr x ;;
      comm c).
Next Obligation.
move=>i [v][h][->]J; apply: vrfD=> D; apply: val_ret; split=>//= k.
move/(_ h v J)=>->; apply: bnd_do=>/=; first by hhauto.
case=> {k} k; move/(_ h v.+1 (erefl _))=>[->]_; apply: val_throw=>/= h1 w1.
by rewrite J; case/(cancel D)=><- _ ->; right; split=>//.
Qed.
Next Obligation.
rewrite !gh; apply: ghE; first by move=> i [_] [v][h]-> _; exists (v,h).
move=> i p -> [->][h][v]E; rewrite E=>D; apply: val_do=>//=; first by hhauto.
move=> m k Dk [->] H2.
have P1: incr3P1 (x :-> v :+ h) x (x :-> v.+1 :+ h) by
  move=> h1 v1 ;case/(cancel D)=>->_->.
(have D1: def (x :-> v.+1 :+ h) by rewrite !defPtUn in D * ); left.
by case: (H2 (x :-> v.+1 :+ h) P1 D1)=>[H3 H4]; split=>//=.
Qed.
Next Obligation.
rewrite !gh; apply: ghE; first by move=> i [v][h]-> _; exists (v,h).
move=> i p -> [v][h]E; rewrite E=> D; apply: bnd_do=>// ; first by hhauto.
move=>/= m k; case/(_ h v (erefl _))=>[[->]H2|[->H2]].
(* First iteration*)
+ apply:bnd_do; first (by hhauto); move=>/= tt {k} k .
 move/(_ h v (erefl _))=>[->]_.
have P1: incr3P1 (x :-> v :+ h) x (x :-> v.+1 :+ h). 
 - by move=> h1 v1 ;move/(cancel D)=>[->]_->.
have D1: def (x :-> v.+1 :+ h) by rewrite !defPtUn in D *.
case: (H2 (x :-> v.+1 :+ h) P1 D1)=>[H3 H4]; apply:val_do=>//= {k tt} tt k Dk.
by move/(H4 tt k); move/(_ Dk); apply: False_ind.
(*jumped back in *)
(* Last iteration *)
apply:bnd_do; first by hhauto.
move=>/= tt{k} k; move/(_ h v.+2 (erefl _))=>[->]_;apply:vrfD=>D1.
case: (H2 (x :-> v.+3 :+ h) _ D1)=>//[H3 H4]. 
apply: val_do=>//= [[]] {k} k Dk.
by case/(H4 _ k)/(_ Dk)=>_ ->; case:(cancel D (esym E))=><-_<-.
Qed.

(* Algebraic Model's proof *)

(* Next Obligation. *)
(* move=> i [v] [h][->][J]; apply: vrfD=>D0; apply: val_ret. *)
(* split=>//= i1; rewrite J; move/(_ h v (erefl _))=>[->]. *)
(* apply: val_abort; split; first by split=>// h2 v2; move/(cancel D0)=>[<-]_->. *)
(* move=>/= k [_ [->]]; apply: bnd_do=>//=; first by hhauto. *)
(* move=> tt k1; move/(_ h (v.+1) (erefl _))=>[->] _. *)
(* apply: val_ret=>// h1 v1; move/(cancel D0)=>[<-]_<- _. *)
(* by split=>//= k2; move/(_ h v (erefl _))=>->; apply: val_ret. *)
(* Qed. *)
(* Next Obligation. *)
(* move => i [v][h][->] ; apply: vrfD=> D0; apply: bnd_cccR. *)
(* + apply: bnd_do=>/= ; first by hhauto. *)
(*  move => c k [->] H2. *)
(*  apply: bnd_do; first by hhauto. *)
(*  move=> {k} tt k; move/(_ h v (erefl _))=>[->]_; apply: vrfD=>D1. *)
(*  have P1: incr3P1 (x :-> v :+ h) x (x :-> v.+1 :+ h). *)
(*  - by move=> h1 v1 ;move/(cancel D0)=>[->]_->. *)
(*  case: (H2 (x :-> v.+1 :+ h) P1 D1)=>[H3 H4].  *)
(*  apply:val_do=>// w k1 Dk1 P. *)
(*  by move:( H4 w k1 P Dk1); apply: False_ind.  *)
(* move=> j; move/(_ h v (erefl _))=>[->]. *)
(* apply: bnd_do=>//= m k; move/(_  h v (erefl _) (erefl _))=>[->]H1. *)
(* apply: vrfD=>D1;apply: bnd_do; first (by hhauto); move=>/= {k} tt k. *)
(* move/(_ h v.+2 (erefl _))=>[->]_. *)
(* have P1: incr3P1 (x :-> v.+2 :+ h) x (x :-> v.+3 :+ h). *)
(*  by move=> h1 v1; move/(cancel D1)=>[->]_[->]. *)
(* apply:vrfD; case/(H1 (x :-> v.+3 :+ h) P1)=>[H3 H4].  *)
(* apply: val_do=>//= {k tt} tt k Dk P. *)
(* by move:( H4 tt k P Dk)=>[<-][->] v1 h1; move/(cancel D0)=>[->][_][->]. *)
(* Qed. *)






Let sf x := (fun i => exists h, exists v : nat, 
                              i = x :-> v  :+ h, 
                fun r i m => forall h (w: nat),
                                i = x :-> w  :+ h ->
                                ( m = i /\ CCC.conseq (spec_of r) 
                             (incr3P1 i x) incr3Q1)
                                \/ 
                                ( m = x :-> w.+2 :+ h  /\
                               CCC.conseq (spec_of r) (incr3s.1) incr3s.2)).


Program
Definition incrV2 (x: ptr) : ST 
  ( fun i =>  exists v : nat, exists h, i = x :-> v :+ h,
     fun y i m =>  forall v h, i = x :-> v :+ h ->
                       m = x :-> v .+3 :+ h /\ y = tt) :=
  Do (c <-- @CCC.callcc _  (sf x) 
        (fun j => ( fun i => j = i /\  (sf x).1 i, (sf x).2))                              
        (fun j ex1 =>
           Do (ret (BigST (incr x ;; throw ex1 (BigST (ret tt)) _)))) _;
      incr x ;;
      comm c).
Next Obligation.
rewrite !gh; apply: ghE; first by move=> i [_][h][v]->_; exists (h, v).
move=>/= i p -> [J][h][v]I; rewrite I in J *=>D; apply: val_ret; left.
split=>//= k; move/(_ h v (erefl _ ))=>->.
apply: bnd_do=>/=; first by hhauto.
case=> {k} k; move/(_ h v.+1 (erefl _))=>[->]_; apply: val_throw=>/= h2 w2.
by rewrite J; case/(cancel D)=><- _ ->; right; split=>//.
Qed.
Next Obligation. by []. Qed.
Next Obligation.
rewrite !gh; apply: ghE; first by move=> i [v][h]-> _; exists (v,h).
move=> i p -> [v][h]E; rewrite E=> D; apply: bnd_do=>// ; first by hhauto.
move=>/= m k; case/(_ h v (erefl _))=>[[->]H2|[->H2]].
(* First iteration*)
+ apply:bnd_do; first (by hhauto); move=>/= tt {k} k .
 case/(_ h v (erefl _))=>->_.
have P1: incr3P1 (x :-> v :+ h) x (x :-> v.+1 :+ h) 
 by move=> h1 v1 ;case/(cancel D)=>->_->.
have D1: def (x :-> v.+1 :+ h) by rewrite !defPtUn in D *.
case: (H2 (x :-> v.+1 :+ h) P1 D1)=>[H3 H4]; apply:val_do=>//= {k tt} tt k Dk.
by move/(H4 tt k); move/(_ Dk); apply: False_ind.
(*jumped back in *)
(* Last iteration *)
apply:bnd_do; first by hhauto.
move=>/= tt{k} k; move/(_ h v.+2 (erefl _))=>[->]_;apply:vrfD=>D1.
case: (H2 (x :-> v.+3 :+ h) _ D1)=>//[H3 H4]. 
apply: val_do=>//= [[]] {k} k Dk.
by case/(H4 _ k)/(_ Dk)=>_ ->; case:(cancel D (esym E))=><-_<-.
Qed.
End Incr3.
