(************************************************)
(* Rules for discharging proof obligations from *)
(*         the ST model for callcc v0.2         *)
(*              WHOLE HEAP                      *)
(************************************************)

From mathcomp Require Import ssreflect ssrfun ssrbool ssrnat eqtype seq. 
From Kont Require Import prelude rels domains heaps hprop callccWH.
Require Import Eqdep.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

(* should also be included here. -- Germ'an *)
Open Scope ccscope.

Section PullDef.
Variables (A : Type) (i : heap).
Variables (q : A -> heap -> Prop) (s: spec A).

Lemma vrfD : (def i -> CCC.verify i s q) -> CCC.verify i s q.
Proof. by move=>H Di; apply:(H Di). Qed.
End PullDef.

Section EvalDo.
Variables (A B : Type) (i : heap).
Variables (q : A -> heap -> Prop) (r : B -> heap -> Prop).
Variables (s1: spec A) (s2 : A -> spec B).

Lemma val_do :
  s1.1 i -> (forall x k, def k -> s1.2 x i k -> q x k) ->
     CCC.verify i s1 q.
Proof. by move => H H1 ; split => // x k Dk H2; apply: (H1 x k H2 Dk). Qed.

Lemma bnd_do :   
   s1.1 i -> (forall x k, s1.2 x i k -> CCC.verify k (s2 x) r) -> 
      CCC.verify i (CCC.bind_s s1 s2) r.
Proof.
move=>H1 H2 D; split=>[|y m].
- by split=>[//|y m] Dm H3; case:(H2 _ _ H3 Dm).
by case=>x[h][Dh][H3] H4 Dm; case:(H2 _ _ H3 Dh)=>H5; apply.
Qed.
End EvalDo.

Section EvalRet.
Variables (A B : Type) (i : heap).
Variables (q : A -> heap -> Prop) (r : B -> heap -> Prop).
Variables (s1: spec A) (s2 : A -> spec B).

Lemma val_ret v:  q v i -> CCC.verify i (CCC.ret_s v) q. 
Proof. by move=>H; apply val_do => // x k Dk [-> ->]. Qed.

Lemma bnd_ret v: 
   CCC.verify i (s2 v) r -> CCC.verify i (CCC.bind_s (CCC.ret_s v) s2) r.
Proof. by move=>H; apply: bnd_do => // x h [-> ->]. Qed.
End EvalRet.

Section EvalBind.
Variables (A B C: Type) (i : heap).
Variables (q : A -> heap -> Prop) (r : C -> heap -> Prop).
Variables (s1: spec A) (s2 : A -> spec B) (s3: B -> spec C).

Lemma bnd_bnd:
  CCC.verify i (CCC.bind_s s1 (fun x => CCC.bind_s (s2 x) s3)) r -> 
        CCC.verify i (CCC.bind_s (CCC.bind_s s1 s2) s3) r. 
Proof.
move=>H ; apply: vrfD => Di. 
case: H => // [H1] H2 ; case: H1 => P1 H3.
apply: bnd_do => // /= [ | y k [x] [h] [Dh] [Q1] Q2].
+ by split => // x m Dm Qm; move: (H3 x m Dm Qm)=> [H4] _.
move => Dk; move: (H3 x h Dh Q1) => [_ H6].
split => [| z m Q3 Dm]; first by apply: (H6 y k Dk Q2).
apply: (H2 z m) => // /=.
by exists x; exists h; repeat split => //; exists y; exists k.
Qed.
End EvalBind.

Section EvalRead.

Variables (A B: Type) (v: A) (x : ptr).
Variables (i h : heap) (q: A -> heap -> Prop) (r : B -> heap -> Prop).
Variables (s : A -> spec B).

Lemma val_read : 
        (def (x:-> v :+ h) -> q v (x :-> v :+ h)) -> 
           CCC.verify (x :-> v :+ h) (CCC.read_s A x) q.
Proof.
move => H; apply val_do => //=; first by exists v;exists h.
move => y k Dk [E H1] ; rewrite E in  Dk *.
by rewrite (H1 v h) //= ; apply: (H Dk).
Qed.

Lemma bnd_read: 
     CCC.verify (x :-> v :+ i) (s v) r -> 
        CCC.verify (x :-> v :+ i) (CCC.bind_s (CCC.read_s A x) s) r.
Proof.
move => H; apply: bnd_do => //= [|y k [-> H2]] ; first by exists v ; exists i.
by move: (H2 v i (erefl (x:-> v :+ i))) ->.
Qed.
End EvalRead.

Section EvalWrite.
Variables (A B C : Type) (v : A) (w : C) (x : ptr).
Variables (i h : heap) (q: unit -> heap -> Prop) (r : B -> heap -> Prop).
Variables (s : unit -> spec B).

Lemma val_write: 
    (def (x :-> v :+ i) -> q tt (x :-> v :+ i)) -> 
        CCC.verify (x :-> w :+ i) (CCC.write_s x v) q.
Proof.
move => H; apply val_do => //=; first by exists C; exists w ;exists i .
move => u k Dk H1.
case:(H1 C w i (erefl (x:-> w :+ i))) =>E ->; rewrite E in Dk *.
by apply: (H Dk).
Qed.

Lemma bnd_write: 
        CCC.verify (x :-> v :+ i) (s tt) r -> 
        CCC.verify (x :-> w :+ i) (CCC.bind_s (CCC.write_s x v) s) r.
Proof.
move=>H.
apply:bnd_do =>//= [|u k H1]; first by exists C; exists w; exists i.
by case:(H1 C w i (erefl (x:-> w :+ i)))=>->->.
Qed.
End EvalWrite.

Section EvalAlloc.
Variables (A B: Type) (v: A) (x : ptr).
Variables (i h : heap) (q: ptr -> heap -> Prop) (r : B -> heap -> Prop).
Variables (s : ptr -> spec B).

Lemma val_alloc : 
        (forall x, def (x :-> v :+ i) -> q x (x :-> v :+ i)) -> 
        CCC.verify i (CCC.alloc_s v) q.
Proof.
move=> H; apply: val_do=>//.
by move=> //= y k Dk H0 ; rewrite H0 in Dk *; apply: H.
Qed.

Lemma bnd_alloc : 
        (forall x, CCC.verify (x :-> v :+ i) (s x) r) ->
        CCC.verify i (CCC.bind_s (CCC.alloc_s v) s) r.
Proof. by move => H ; apply: bnd_do => //= y m ->. Qed.
End EvalAlloc.

Section EvalDealloc.
Variables (A B : Type) (x: ptr) (v: A).
Variables (i : heap) (q: B -> heap -> Prop) (r : unit -> heap -> Prop).
Variables (s : unit -> spec B).

Lemma val_dealloc: 
        (def i -> r tt i) -> 
        CCC.verify (x :-> v :+ i) (CCC.dealloc_s x) r.
Proof.
move=> H; apply: vrfD => H0.
apply: val_do => //= ; first by [exists A; exists v; exists i].
move=>y m Dm H1; rewrite H1 freePtUn in Dm *; last (by []).
by elim y; apply: H.
Qed.

Lemma bnd_dealloc: 
        CCC.verify i (s tt) q -> 
        CCC.verify (x :-> v :+ i) (CCC.bind_s (CCC.dealloc_s x) s) q.
Proof. 
move=> H ; apply: vrfD => H0.
apply:bnd_do => //= ; first by [exists A; exists v ; exists i].
by move => y k ->; rewrite freePtUn; elim y.
Qed.

End EvalDealloc.

Section EvalCC.
Variables (A B: Type) (p: A -> pre) (s1 : spec A).
Variables (i : heap) (q: B -> heap -> Prop) (r : A -> heap -> Prop).
Variables (v:A) (s2 : A -> spec B).

Lemma val_cc:
  s1.1 i ->
     (forall x k, def k -> s1.2 x i k -> r x k) ->
        CCC.verify i (CCC.callcc_s s1) r.
Proof. by move => S H;apply:val_do => //=. Qed.

(* val_cc is just a specialized val_do *)

Lemma bnd_cc:
  s1.1 i ->
    (forall y k, s1.2 y i k  -> CCC.verify k (s2 y) q) ->
      CCC.verify i (CCC.bind_s (CCC.callcc_s s1) s2) q.
Proof. by move => H0 H; apply: bnd_do => //=. Qed.

(* Analogously, this is just bnd_do                               *)
(* It is not possible to deduce automatically further information *)
(* w.o having a specific rule for each combination                *)
(* e.g. a bnd_read_call rule and resp. a bnd_callcc_read rule     *)

End EvalCC.

Section EvalThrow.
Variables (A B C: Type) (p : A -> heap -> Prop) (f : STcont p) (x : A).
Variables (i : heap) (s2: B -> spec C) .
Variables ( r: B -> heap -> Prop) (q: C -> heap -> Prop).
(* Definition throw_s B : spec B := (p x, fun y i m => False). *)
Lemma bnd_throw: p x i -> 
        CCC.verify i (CCC.bind_s (CCC.throw_s p x B) s2) q.
Proof. by move=> H Di; split=> // /= y m [z] [h] [H1] [F] H2 Dm. Qed.

Lemma val_throw: p x i -> 
        CCC.verify i (CCC.throw_s p x B) r.
Proof. by move=> H Di; split=> // /= y m [z] [h] [H1] [F] H2 Dm. Qed.


(* This lemmas are too weak, we need to be able to relate the heaps *)
(* Not sure on how, and which, and why *)
(* Maybe we could use some sort of simplification rules (i.e. rewriteing)  *)
(* for specs involving throw_s *)

(* The following lemma implements the monadic property: *)
(* throw k x >>= m =  throw k x *)  
(* in the ST idiom *)
 
Lemma throw_cancel : 
  CCC.verify i (CCC.throw_s p x C) q ->
    CCC.verify i (CCC.bind_s (CCC.throw_s p x B) s2) q.
Proof.
move=> H Di /=.
case: H=> //= P0 H1.
by split=> [| y m [w][t][Dt][F H] Dm]; first by split=>// x m Dm F.
Qed.

(* However this lemma not much useful, as it is weaker than bind_throw *)

End EvalThrow.

(******************************************)
(* Lemmas for pulling out ghost variables *)
(******************************************)

Section Ghosts.
Variables (A : Type) (s : spec A) (p : pre).

Lemma allC (B1 B2 : Type) (q : B1 -> B2 -> post A) :
       CCC.conseq s p (fun y i m => forall x1 x2, q x1 x2 y i m) <->
       CCC.conseq s p (fun y i m => forall x, q x.1 x.2 y i m).
Proof.
split=>//= H1 i H2 Di; case: (H1 i H2 Di) => S1 H3; split => //= y m S2 Dm.
- by move => [x1 x2]; apply: H3.
by move => x1 x2; apply: (H3 y m S2 Dm (x1,x2)).
Qed. 
   
Lemma impC (B : Type) (q1 q2 : heap -> B -> Prop) (r : B -> post A) :
        CCC.conseq s p (fun y i m => forall x, q1 i x -> q2 i x -> r x y i m) <->
        CCC.conseq s p (fun y i m => forall x, q1 i x /\ q2 i x -> r x y i m).
Proof.
split => //= H1 i H2 Di; case: (H1 i H2 Di) => S1 H3; split => //= y m S2 Dm x.
- by case => H4 H5 ; apply: (H3 y m S2 Dm).
by move => H4 H5 ; apply: (H3 y m S2 Dm).
Qed.

Lemma ghE (B : Type) (q : heap -> B -> Prop) (r : B -> post A) : 
        (forall i, p i -> def i -> exists x, q i x) ->
        (forall i x, q i x -> p i -> def i ->
           CCC.verify i s (fun y m => r x y i m)) ->
        CCC.conseq s p (fun y i m => forall x, q i x -> r x y i m).
Proof.
move=>H1 H2 i H3 Di; case: (H1 i H3 Di)=>x H4.
case: (H2 i x H4 H3 Di Di)=>H5 _; split=>// y m H6 Dm z H7.
by case: (H2 i z H7 H3 Di Di)=>_; apply.
Qed.

End Ghosts.

Definition gh := (allC, impC).